//  @flow
import express from 'express';
import webpack from 'webpack';
import webpackDevMiddleware from 'webpack-dev-middleware';
import webpackHotMiddleware from 'webpack-hot-middleware';
import path from 'path';

import type { $Request, $Response } from 'express';

import { getMessage } from './utils/messaging/messaging';
import getWebpackConfig from './webpack/webpack.config';
import composedMiddleware from './middleware';
import hooks from './hooks';
import defaultConfig from './webpack/defaultConfig';

import type { IConfig } from './../flow/IConfig';
import type { IDiagnosticRoutes } from './../flow/IDiagnosticRoutes';

//  Run hooks before
//  starting application
hooks();

export const createServer = async (
  config: IConfig,
  diagnosticRoutes: IDiagnosticRoutes,
  callback: ?Function,
) => {
  const {
    appName,
    port,
    publicPath = defaultConfig.publicPath,
    distOutput = defaultConfig.distOutput,
    isDevEnvironment,
  } = config;

  const app = express();

  //  Let express.static handle all static assets
  if (!isDevEnvironment) {
    app.use(publicPath, express.static(path.join(process.cwd(), distOutput, publicPath)));
  }

  //  Inject webpack-dev-middleware and hot-reloading
  if (isDevEnvironment) {
    const webpackConfig = getWebpackConfig(config);
    const compiler = webpack(webpackConfig);
    const hotMiddleware = webpackHotMiddleware(compiler);

    //  Sets the compiler locally in the express cache
    app.locals.compiler = compiler;
    app.locals.hotMiddleware = hotMiddleware;

    app.use(publicPath, webpackDevMiddleware(compiler));
    app.use(hotMiddleware);
  }

  //  Apply diagnostic routes
  if (diagnosticRoutes) {
    Object.keys(diagnosticRoutes).forEach(key => (
      app.get(key, diagnosticRoutes[key])
    ));
  }

  //  Apply composed middleware to each request
  app.use(composedMiddleware(app, config), (req: $Request, res: $Response) => (
    res.send(app.locals.output)),
  );

  //  Start on port
  app.listen(port || 3000, () => {
    if (callback && typeof callback === 'function') {
      callback(getMessage('welcome', { port, appName, isDevEnvironment }));
    }
  });

  return app;
};

export default createServer;
